/*
 * Vorlesung ``Data Processing on Modern Hardware''
 *
 * Uebungsblatt 2/3
 *
 * Diesen Code werden wir erst im folgenden Aufgabenblatt
 * verwenden; er basiert auf Code, der von Rene Mueller an
 * der ETH Zuerich entwickelt wurde.
 */

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include "engine.h"


/*
 * implementation of operators 
 */


/* produce next output tuples for aggregation operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int 
next_aggop(int32_t *tuples, aggop_t *op)
{
   uint32_t r;
   int tuple;
   for (;;) {
      r = op->input->next(&tuple, op->input);
      if (r > 0) {
         /* valid tuple returned */
         if (__builtin_expect(op->count == 0, 0)) { /* unlikely */
            /* first tuple */
            op->extremum = tuple;
         }
         op->count++;
         op->sum += tuple;
         if ((op->aggregate == MAX) && 
               (tuple > op->extremum)) {
            op->extremum = tuple;
         }
         if ((op->aggregate == MIN) && 
               (tuple < op->extremum)) {
            op->extremum = tuple;
         }
      } else if (r == 0) {
         /* end of scan: compute aggregate */
         if (op->count == 0) {
            return 0;
         } else {
            switch (op->aggregate) {
            case AVG:
               *tuples = op->sum/op->count;
               break;
            case COUNT:
               *tuples = op->count;
               break;
            case MAX:
            case MIN:
               *tuples = op->extremum;
               break;
            case SUM:
               *tuples = op->sum;
               break;
            default:
               fprintf(stderr, "invalid aggregate %d\n", op->aggregate);
               return -1;
            }
            return 1;
         }
      } else {
         /* error condition */
         return -1;
      }
   }
}

/* produce next output tuples for arithmetic operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int 
next_arithop(int32_t *tuples, arithop_t *op)
{
   int32_t left, right; 
   int numleft, numright;
   
   numleft = op->leftop->next(&left, op->leftop);
   if (numleft == -1) {
      return -1;  /* propagate error */
   } else if (numleft == 0) {
      /* no tuples: treat it as "pseudo" NULL */
      left = 0;
   }

   numright = op->rightop->next(&right, op->rightop);
   if (numright == -1) {
      return -1;  /* propagate error */
   } else if (numright == 0) {
      /* no tuple: treat it as "pseudo" NULL */
      right = 0;
   }
   
   if ((numleft == 0) && (numright == 0)) {
      /* no tuples from both sides -> end of relation */
      return 0;
   }

   switch (op->operation) {
   case ADD:
      *tuples = left+right;
      break;
   case SUB:
      *tuples = left-right;
      break;
   case MULT:
      *tuples = left*right;
      break;
   case DIV:
      *tuples = left/right;
      break;
   default:
      fprintf(stderr, "invalid arith operation: %d\n", op->operation);
      return -1;
   }   
   return 1;
}


/* produce next output tuples for scan operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int 
next_scanop(void *tuples, scanop_t *op)
{
   uint8_t *bytes;

   if (op->row < op->numrows) {
      bytes = ((uint8_t*)op->rows) + op->row*op->rowwidth + op->offset;  
   
      switch (op->width) {
      case 1:
         *((uint8_t*)tuples) = *bytes;
         break;
      case 4:    
         *((uint32_t*)tuples) = *((uint32_t*)bytes); 
         break; 
      case 8:
         *((uint64_t*)tuples) = *((uint64_t*)bytes);
         break;
      default:
         fprintf(stderr, "invalid scan width: %d\n", op->width);
         return -1;    
      }
      op->row++;
      return 1;
   } else {
      /* scan reached end of relation */
      return 0;
   }
}


/*
 * allocation, reset and deallocation functions for operators  
 */

/* allocate and initialize aggregation operator 
 * \param op      aggregation operator object
 * \param agg     aggregate function 
 * \param input   operator producting input
 * \return 0 upon success, -1 otherwhise
 */
int 
init_aggop(aggop_t *op, aggregate_t agg, anyop_t* input)
{
   op->next = next_aggop;
   op->input = input; 
   op->aggregate = agg;
   op->count = 0;
   op->sum = 0;
   return 0;
}

/* allocate and initialize arithmetic operator 
 * \param op      arithmetic operator object
 * \param o       arithmetic operation (ADD, SUB, MULT, DIV)     
 * \param leftop  operator producting data for left-side
 * \param rightop operator producting data for right-side
 * \return 0 upon success, -1 otherwhise
 */
int 
init_arithop(arithop_t *op, operation_t o, anyop_t *leftop, anyop_t *rightop)
{
   op->operation = o;
   op->next = next_arithop; 
   op->leftop = leftop; 
   op->rightop = rightop; 
   return 0;
}

/* allocate and initialize scan operator 
 * \param op         scan operator object
 * \param relation   pointer to base relation
 * \param numrows    number of rows in base relation
 * \param rowwidth   width of a single row (in bytes)
 * \param rowoffset  offset of field in a single row (in bytes)
 * \param readwidth  read length (in bytes) 
 * \return 0 upon success, -1 otherwhise
 */
int 
init_scanop(scanop_t *op, void *relation, 
            uint32_t numrows, uint32_t rowwidth, 
            uint32_t rowoffset, uint32_t readwidth)
{
   op->next = next_scanop; 
   op->rows = relation; 
   op->numrows = numrows; 
   op->rowwidth = rowwidth; 
   op->offset = rowoffset;
   op->width = readwidth;
   op->row = 0; 
   return 0;
}

/* reset aggregation operator */
void 
reset_aggop(aggop_t *op)
{
   op->count = 0;
   op->sum = 0;
}


/* reset state of arithmetic operator */ 
void 
reset_arithop(arithop_t *op) 
{
   /* nothing to do */
}


/* reset state of scan operator */
void 
reset_scanop(scanop_t *op)
{
   op->row = 0;
}

/* free (deallocate) aggregation operator */
void 
free_aggop(aggop_t *op)
{
   /* nothing to be done (yet) */
}

/* free (deallocate) arithmetic operator */
void 
free_arithop(arithop_t *op)
{
   /* nothing to be done (yet) */
}

/* free scan operator */
void 
free_scanop(scanop_t *op)
{
   /* nothing to be done */
}

