/*
 * Vorlesung ``Data Processing on Modern Hardware''
 *
 * Uebungsblatt 2/3
 *
 * Diesen Code werden wir erst im folgenden Aufgabenblatt
 * verwenden; er basiert auf Code, der von Rene Mueller an
 * der ETH Zuerich entwickelt wurde.
 */

 
#ifndef ENGINE_H
#define ENGINE_H


/* general operator (base type) */
typedef struct anyop {
   /* function pointer to operator implementation:
    * return value  > 0  corresponds to the number of tuples returned  
    *              == 0  means end of relation reached 
    *              == -1 means error occurred */
   int         (*next)(void *tuples, struct anyop *op);
} anyop_t;

/* aggregate function used in aggregation operator */
typedef enum { AVG, COUNT, MAX, MIN, SUM } aggregate_t;


/* aggregation operator */
typedef struct aggop {
   /* function pointer to operator implementation:
    * return value  > 0  corresponds to the number of tuples returned  
    *              == 0  means end of relation reached 
    *              == -1 means error occurred */
   int         (*next)(int32_t *tuples, struct aggop *op); 

   /* operator parameters */
   aggregate_t aggregate;  /* aggregate computed by this operator */
   anyop_t     *input;     /* operator producing input */

   /* operator state */
   int32_t count;          /* counts number of tuples */ 
   int32_t sum;            /* sums up tuple values */
   int32_t extremum;       /* exemplary value (e.g., MIN or MAX) */
} aggop_t;


/* operation implemented in arithmetic operator */
typedef enum { ADD, SUB, MULT, DIV } operation_t;


/* arithmetic operator of the form   
 *          (left suboperator) op (right suboperator)
 */
typedef struct arithop {
   /* function pointer to operator implementation:
    * return value  > 0  corresponds to the number of tuples returned  
    *              == 0  means end of relation reached 
    *              == -1 means error occurred */
   int         (*next)(int32_t *tuples, struct arithop *op); 

   /* operator parameters */
   operation_t operation;  /* operation implemented in arithmetic operator */
   anyop_t     *leftop;    /* left child operator */
   anyop_t     *rightop;   /* right child operator */
} arithop_t;


/* scan operator: reads inputs from base relation
 */
typedef struct scanop {
   /* function pointer to operator implementation:
    * return value  > 0  corresponds to the number of tuples returned  
    *              == 0  means end of relation reached 
    *              == -1 means error occurred */
   int         (*next)(void *tuples, struct scanop *op);

   /* operator parameters */
   void        *rows;      /* pointer to base relation */
   uint32_t    numrows;    /* number of rows in base relation */
   uint32_t    rowwidth;   /* width of row in bytes */ 
   uint32_t    offset;     /* offset of field to read from the beginning of the row */
   uint32_t    width;      /* number of bytes to read (1, 4, or 8) */

   /* operator state */
   uint32_t    row;        /* current row */
} scanop_t;


/*
 * operator allocation, initialization, and dealloction functions 
 */

/* allocate and initialize aggregation operator 
 * \param op      aggregation operator object
 * \param agg     aggregate function 
 * \param input   operator producting input
 * \return 0 upon success, -1 otherwhise
 */
int init_aggop(aggop_t *op, aggregate_t agg, anyop_t* input);

/* allocate and initialize arithmetic operator 
 * \param op      arithmetic operator object
 * \param o       arithmetic operation (ADD, SUB, MULT, DIV)     
 * \param leftop  operator producting data for left-side
 * \param rightop operator producting data for right-side
 * \return 0 upon success, -1 otherwhise
 */
int init_arithop(arithop_t *op, operation_t o, 
                 anyop_t *leftop, anyop_t *rightop);

/* allocate and initialize scan operator 
 * \param op         scan operator object
 * \param relation   pointer to base relation
 * \param numrows    number of rows in base relation
 * \param rowwidth   width of a single row (in bytes)
 * \param rowoffset  offset of field in a single row (in bytes)
 * \param readwidth  read length (in bytes) 
 * \return 0 upon success, -1 otherwhise
 */
int init_scanop(scanop_t *op, void *relation, 
            uint32_t numrows, uint32_t rowwidth, 
            uint32_t rowoffset, uint32_t readwidth);

/* reset aggregation operator */
void reset_aggop(aggop_t *op);

/* reset state of arithmetic operator */ 
void reset_arithop(arithop_t *op);

/* reset state of scan operator */
void reset_scanop(scanop_t *op);

/* free (deallocate) aggregation operator */
void free_aggop(aggop_t *op);

/* free (deallocate) arithmetic operator */
void free_arithop(arithop_t *op);

/* free scan operator */
void free_scanop(scanop_t *op);

/* 
 * operator implementations 
 */

/* produce next output tuples for aggregation operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int next_aggop(int32_t *tuples, aggop_t *op);


/* produce next output tuples for arithmetic operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int next_arithop(int32_t *tuples, arithop_t *op);


/* produce next output tuples for scan operator 
 * \param tuples  pointer to buffer where output tuples are to be placed 
 *                (needs to be be allocated outside) 
 * \param op      pointer to aggregation operator struct 
 *                (holds state and pararameters, similar to a 'this' pointer)
 * \return 0: end of releation reached, -1: error occurred, otherwhise the number
 *            of tuples written to buffer 'tuples'
 */
int next_scanop(void *tuples, scanop_t *op);

#endif // ENGINE_H

